/*
 *  vectex - a SVG vector texture plugin for Blender
 *  Copyright (C) 2008 Marcelo de Gomensoro Malheiros <mgmalheiros@gmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>

#include "math.h"
#include "plugin.h"
#include "util.h"

#include "vectex.h"


/*---------------- Main declarations ----------------*/

#define TILE_SIZE       256
#define TILE_SIZE_POWER 8
#define TILE_BORDER     1
#define MAX_TIMESTAMP   5000

/* Special tile age values */
enum {UNUSED = 0, CONST_COLOR, START_VALUE};

typedef struct Tile
{
    unsigned char *image;
    unsigned int   age;
} Tile;

typedef struct Instance
{
    void         *backend_data;
    int           tile_set_max_power;
    Tile         *tile_set;
    int           tile_memory;
    unsigned int  timestamp;
    unsigned char current_base_color[4];
    char          current_file_name[64];
} Instance;


/*---------------- Plugin management declarations ----------------*/

char name[] = "vectex";

/* Names must be less than 15 characters */
#define NR_TYPES 1
char stnames[NR_TYPES][16] = {"Extend"};

/* varstr and Cast must have the same variables in the same order */ 
#define TEX (6<<9)
VarStruct varstr[] =
{
    {TOG|INT, "Enable",       1, 0,   1, "Enable vector texture (disabling frees memory)"},
    {TOG|INT, "MipMap",       1, 0,   1, "Uses mipmap averaging"},
    {TOG|INT, "Interpol",     1, 0,   1, "Interpolates nearby pixels"},
    /*{TOG|INT, "UseAlpha",     0, 0,   1, "Uses texture alpha channel"},*/
    {LABEL,   "",             0, 0,   0, ""},
    {TOG|INT, "Tex Level",    0, 0,   1, "Displays texture detail as hue"},
    {NUM|INT, "Mem Max",     20, 0, 200, "Maximum memory to use (in megabytes, 0=no limit)"},
    {NUM|INT, "Mem:",         0, 0, 999, "Displays current memory being used (in megabytes)"},
    {LABEL,   "Base Color",   0, 0,   0, "Base texture color"},
    {COL|CHA, "",             0, 0, 255, "Base texture color"},
    {LABEL,   "File Name",    0, 0,   0, "SVG file name"},
    {TEX|CHA, "",             0, 0,  63, "SVG file name"},
    {LABEL,   "vectex 0.9",   0, 0,   0, "Plugin name and version"},
};

/* Cast must have max size of 32 floats = 128 bytes and all fields be 32-bit aligned */ 
typedef struct Cast
{
    int           enable;
    int           mipmap;
    int           interpol;
    /*int           usealpha;*/
    int           label_dummy0;
    int           tex_level;
    int           mem_max;
    int           mem;
    int           label_dummy1;
    unsigned char base_color[4];
    int           label_dummy2;
    char          file_name[64];
    int           label_dummy3;
    void         *instance_data;
} Cast;

/* Intensity, R, G, B, Alpha, nor.x, nor.y, nor.z */
float result[8];

/* Current frame number */
float cfra;

void plugin_init(void);
void plugin_callback(int);
void plugin_instance_init(Cast *);
int plugin_tex_doit(int, Cast *, float *, float *, float *, float *);


/*---------------- Plugin management functions ----------------*/

int plugin_tex_getversion(void) 
{    
    MSG0("\n\n[plugin_tex_getversion]\n");
    
    return (B_PLUGIN_VERSION);
}

void plugin_getinfo(PluginInfo *info)
{
    MSG0("[plugin_getinfo]\n");

    info->name   = name;
    info->stypes = NR_TYPES;
    info->snames = stnames[0];
    info->varstr = varstr;
    info->nvars  = sizeof(varstr)/sizeof(VarStruct);

    info->result = result;
    info->cfra   = &cfra;

    info->init          = plugin_init;
    info->callback      = plugin_callback;
    info->instance_init = (void (*)(void *)) plugin_instance_init;
    info->tex_doit      = (TexDoit) plugin_tex_doit;
}

void plugin_init(void)
{
    MSG0("[plugin_init]\n");
}

void plugin_callback(int button) 
{
    MSG0("[plugin_callback]\n");
    
    button = 0; /* prevent unused variable warning */
}


/*---------------- Utility functions ----------------*/

void hsv2rgb(double h, double s, double v, double *pr, double *pg, double *pb)
{
    if (v == 0.0)
    {
        *pr = *pg = *pb = 0.0;
    }
    else if (s == 0.0)
    {
        *pr = *pg = *pb = v;
    }
    else
    {
        double hf = h / 60.0;
        int    i  = floor(hf);
        double f  = hf - i;
        double pv = v * (1 - s);
        double qv = v * (1 - s * f);
        double tv = v * (1 - s * (1 - f));
        switch (i % 6)
        {
          case 0:
            *pr = v;
            *pg = tv;
            *pb = pv;
            break;
          case 1:
            *pr = qv;
            *pg = v;
            *pb = pv;
            break;
          case 2:
            *pr = pv;
            *pg = v;
            *pb = tv;
            break;
          case 3:
            *pr = pv;
            *pg = qv;
            *pb = v;
            break;
          case 4:
            *pr = tv;
            *pg = pv;
            *pb = v;
            break;
          case 5:
            *pr = v;
            *pg = pv;
            *pb = qv;
            break;
        }
    }
}

void *memory_alloc(int len, char *str)
{
    return (mallocN(len, str));
}

void memory_free(void *ptr)
{
    freeN(ptr);
}


/*---------------- Main functions ----------------*/

int get_base_index(int power)
{
    int index = MIN2(TILE_SIZE_POWER, power);    
    int ntiles = 1;
    while (power > TILE_SIZE_POWER)
    {
        index += ntiles;
        ntiles *= 4;
        power--; 
    }
    return (index);
}

void create_tile_set(Instance *instance)
{
    MSG0("[create_tile_set]\n");
    
    instance->tile_set_max_power = TILE_SIZE_POWER;
    int i, size = get_base_index(instance->tile_set_max_power + 1);
    instance->tile_set = (Tile *) memory_alloc(sizeof(Tile) * size, "vectex:tileset");
    for (i = 0; i < size; i++)
    {
        instance->tile_set[i].image = NULL;
        instance->tile_set[i].age   = UNUSED;
    }
    instance->tile_memory = 0;           
    instance->timestamp = START_VALUE;
    
    MSG2("- created tile set max_pow=%d size=%d\n", instance->tile_set_max_power, size);
}

void limit_tile_set(Instance *instance, int maximum)
{
    MSG0("[limit_tile_set]\n");

    int c = 0, i, s = 0, size = get_base_index(instance->tile_set_max_power + 1);
    unsigned int min_age = 0, index = 0;
    while (instance->tile_memory > maximum)
    {
        /* Find oldest tile */
        for (c = i = s = 0; i < size; i++)
        {
            if (instance->tile_set[i].age >= START_VALUE)
            {
                if (c == 0 || instance->tile_set[i].age < min_age)
                { 
                    min_age = instance->tile_set[i].age;
                    index = i;
                }
                c++;
            }
            else if (instance->tile_set[i].age == CONST_COLOR)
            {
                s++;
            }
        }
        
        MSG1("- deleting tile #%d\n", index);
        memory_free(instance->tile_set[index].image);
        instance->tile_set[index].image = NULL;
        instance->tile_set[index].age   = UNUSED;
        
        int tile_size = (index < TILE_SIZE_POWER) ? 1 << index : TILE_SIZE; 
        instance->tile_memory -= (tile_size + TILE_BORDER) * (tile_size + TILE_BORDER) * 3;
    }

    MSG4("- memory now=%d (%.1f Mb) tiles=%d solid=%d\n", instance->tile_memory, 
         instance->tile_memory / 1048576.0, c - 1, s);

    if (instance->timestamp > MAX_TIMESTAMP)
    {
        MSG1("- updating timestamps oldest=%d\n", min_age);
        instance->timestamp -= min_age - START_VALUE;
        for (i = 0; i < size; i++)
        {
            if (instance->tile_set[i].age >= START_VALUE)
            {
                instance->tile_set[i].age -= min_age - START_VALUE;
            }
        }
    }
}

void destroy_tile_set(Instance *instance)
{
    MSG0("[destroy_tile_set]\n");

    int i, size = get_base_index(instance->tile_set_max_power + 1);
    for (i = 0; i < size; i++)
    {
        if (instance->tile_set[i].image != NULL)
        {
            memory_free(instance->tile_set[i].image);
            instance->tile_set[i].image = NULL;
        }
    }
    
    memory_free(instance->tile_set);
    instance->tile_set_max_power = 0;
    instance->tile_set = NULL;
    instance->tile_memory = 0;
    instance->timestamp = 0;
}

void create_instance(Cast *cast)
{
    MSG0("[create_instance]\n");

    Instance *instance = (Instance *) memory_alloc(sizeof(Instance), "vectex:instance");
    cast->instance_data = instance;
    
    void *backend_data = create_backend(cast->file_name);
    if (backend_data == NULL)
    {
        /* Cannot open file, put plugin into error state */
        MSG1("- open failed for file '%s'\n", cast->file_name);
        instance->backend_data = NULL;
        instance->tile_set_max_power = 0;
        instance->tile_set = NULL;
        instance->tile_memory = 0;
        instance->timestamp = 0;
    }
    else
    {
        /* File correctly opened, store backend and create tile set */ 
        MSG0("- open ok\n");
        instance->backend_data = backend_data;
        create_tile_set(instance);
    }
    
    /* Save current color and file name for later comparison */
    memcpy(instance->current_base_color, cast->base_color, 3); 
    strcpy(instance->current_file_name, cast->file_name);
}

void destroy_instance(Cast *cast)
{
    MSG0("[destroy_instance]\n");
    
    Instance *instance = (Instance *) cast->instance_data;
    if (instance->backend_data != NULL)
    {
        /* Cleanup backend data and tile set */
        destroy_backend(instance->backend_data);
        destroy_tile_set(instance);
    }        

    /* Cleanup instance data */ 
    memory_free(cast->instance_data);
    cast->instance_data = NULL;
}

void plugin_instance_init(Cast *cast)
{
    MSG0("[plugin_instance_init]\n");
       
    if (cast->enable)
    {
        /* Plugin is enabled, so create instance data */
        MSG0("- plugin started enabled\n");
        create_instance(cast);
    }
    else
    {
        /* Called by a blender saved file; initialize to disabled state */
        MSG0("- plugin started disabled\n");
        cast->instance_data = NULL;
    }
}

int get_tile(Instance *instance, int x, int y, int power)
{
    int tile_row, tile_col, index;
    tile_row = y / TILE_SIZE;
    tile_col = x / TILE_SIZE;

    if (power > instance->tile_set_max_power)
    {
        /* Create a new, bigger tile set */
        int new_size = get_base_index(power + 1);
        Tile *new_set = (Tile *)memory_alloc(sizeof(Tile) * new_size, "vectex:newtileset");
        
        /* Copy contents from old tile set and fill the remaining space with NULLs */
        int i, old_size = get_base_index(instance->tile_set_max_power + 1);
        memcpy(new_set, instance->tile_set, sizeof(Tile) * old_size);
        for (i = old_size; i < new_size; i++)
        {
            new_set[i].image = NULL;
            new_set[i].age   = UNUSED;
        }
        
        /* Free old tile set and update global variables */
        memory_free(instance->tile_set);
        instance->tile_set_max_power = power;
        instance->tile_set = new_set;
        MSG3("- growing tile set max_pow=%d size=%d old_size=%d\n",
             instance->tile_set_max_power, new_size, old_size);
    }

    if (power <= TILE_SIZE_POWER)
    {
        /* The image is made of just a single tile for this resolution */
        index = get_base_index(power);
    }
    else
    {
        /* The image is made of several tiles for this resolution */     
        int stride = 1 << (power - TILE_SIZE_POWER);
        index = get_base_index(power) + stride * tile_row + tile_col;
    }

    if (instance->tile_set[index].image != NULL)
    {
        /* The needed tile is available, so just increase its timestamp */
        if (instance->tile_set[index].age != CONST_COLOR)
        {
            instance->tile_set[index].age = instance->timestamp++;
        }
    }
    else
    {
        /* The needed tile is not yet available, so render it */
        int resolution = 1 << power;
        int tile_size  = (power < TILE_SIZE_POWER) ? resolution : TILE_SIZE; 

        MSG4("> creating tile #%d res=%d [%d,%d]\n", index, resolution, tile_row, tile_col);
        unsigned char *image = backend_render_tile(instance->backend_data, resolution,
                                                   tile_size, tile_row, tile_col,
                                                   instance->current_base_color[0] / 255.0,
                                                   instance->current_base_color[1] / 255.0,
                                                   instance->current_base_color[2] / 255.0,
                                                   TILE_BORDER);
        
        /* Check if the tile is made of a constant color */
        int i, size = (tile_size + TILE_BORDER) * (tile_size + TILE_BORDER) * 3;
        for (i = 3; i < size; i += 3)
        {
            if (image[0] != image[i] || image[1] != image[i + 1] ||
                image[2] != image[i + 2])
            {
                break;
            }
        }
        if (i == size)
        {
            /* Create a special tile just for this color and free the other one */
            MSG3(">> tile is of solid color #%02x%02x%02x\n", image[0], image[1], image[2]);
            unsigned char *color = (unsigned char *)memory_alloc(3, "vectex:color");
            memcpy(color, image, 3);
            memory_free(image);
            instance->tile_set[index].image = color;
            instance->tile_set[index].age = CONST_COLOR;
            instance->tile_memory += 3; 
        }
        else
        {
            instance->tile_set[index].image = image;
            instance->tile_set[index].age = instance->timestamp++;
            instance->tile_memory += size;
        }
    }

    return (index);
}

void sample(Instance *instance, float *texvec, int power, int interp,
            double *pr, double *pg, double *pb)
{
    int resolution = 1 << power;
    int tile_size  = (power < TILE_SIZE_POWER) ? resolution : TILE_SIZE; 
    
    if (interp == 0)
    {
        /* Simply pick and return the nearest pixel */
        int x = (int)((texvec[0] + 1.0) * (resolution >> 1));
        int y = (int)((1.0 - texvec[1]) * (resolution >> 1));
        
        /* extend: copy the color of edge pixels */
        x = CLAMP(x, 0, resolution - 1);
        y = CLAMP(y, 0, resolution - 1);
        /* clip: alpha=0 outside edges */
        /* repeat: duplicate horizontally and vertically */
        
        lock_thread(LOCK_IMAGE);
        int index = get_tile(instance, x, y, power);
        unlock_thread(LOCK_IMAGE);

        unsigned char *tile = instance->tile_set[index].image;
        if (instance->tile_set[index].age == CONST_COLOR)
        {    
            /* Constant color tile */
            *pr = (double)(tile[0] / 255.0);
            *pg = (double)(tile[1] / 255.0);
            *pb = (double)(tile[2] / 255.0);
        }
        else
        {
            /* Clip coordinates to be inside the tile */
            x %= tile_size;
            y %= tile_size;

            /* Calculate pixel offset taking the border into account */
            int offset = (y * (tile_size + TILE_BORDER) + x) * 3;
            *pr = (double)(tile[offset]     / 255.0);
            *pg = (double)(tile[offset + 1] / 255.0);
            *pb = (double)(tile[offset + 2] / 255.0);
        }
    }
    else
    {
        /* Pick the four nearest pixels and return a bilinear interpolation */
        double u = (texvec[0] + 1.0) * (resolution >> 1) - 0.5;
        double v = (1.0 - texvec[1]) * (resolution >> 1) - 0.5;
        int x = (int)u;
        int y = (int)v;
        double u_ratio = u - x;
        double v_ratio = v - y;
        double u_opposite = 1 - u_ratio;
        double v_opposite = 1 - v_ratio;
        
        /* extend: copy the color of edge pixels */
        x = CLAMP(x, 0, resolution - 1);
        y = CLAMP(y, 0, resolution - 1);
        /* clip: alpha=0 outside edges */
        /* repeat: duplicate horizontally and vertically */

	lock_thread(LOCK_IMAGE);
        int index = get_tile(instance, x, y, power);
        unlock_thread(LOCK_IMAGE);

        unsigned char *tile = instance->tile_set[index].image;
        if (instance->tile_set[index].age == CONST_COLOR)
        {    
            /* Constant color tile */
            *pr = (double)(tile[0] / 255.0);
            *pg = (double)(tile[1] / 255.0);
            *pb = (double)(tile[2] / 255.0);
        }        
        else
        {
            /* Clip coordinates to be inside the tile */
            x %= tile_size;
            y %= tile_size;
            int x_next = x + 1;
            int y_next = y + 1;        
    
            /* Calculate pixel offset taking the border into account */
            int tl = (y      * (tile_size + TILE_BORDER) + x)      * 3; /* top left */
            int tr = (y      * (tile_size + TILE_BORDER) + x_next) * 3; /* top right */
            int bl = (y_next * (tile_size + TILE_BORDER) + x)      * 3; /* bottom left */
            int br = (y_next * (tile_size + TILE_BORDER) + x_next) * 3; /* bottom right */
            *pr = ((tile[tl]     * u_opposite + tile[tr]     * u_ratio) * v_opposite + 
                   (tile[bl]     * u_opposite + tile[br]     * u_ratio) * v_ratio) / 255.0;    
            *pg = ((tile[tl + 1] * u_opposite + tile[tr + 1] * u_ratio) * v_opposite + 
                   (tile[bl + 1] * u_opposite + tile[br + 1] * u_ratio) * v_ratio) / 255.0;    
            *pb = ((tile[tl + 2] * u_opposite + tile[tr + 2] * u_ratio) * v_opposite + 
                   (tile[bl + 2] * u_opposite + tile[br + 2] * u_ratio) * v_ratio) / 255.0;
        }
    }
}

int plugin_tex_doit(int stype, Cast *cast, float *texvec, float *dxt, float *dyt, float *result)
{
    stype = 0; /* prevent unused variable warning */

    if (! cast->enable)
    {
        /* Plugin is disabled, check if there is allocated memory to cleanup */
        if (cast->instance_data != NULL)
        {
            destroy_instance(cast);
            cast->mem = 0;
        }
        
        /* Return a black pixel */
        result[0] = result[4] = 1.0;
        result[1] = result[2] = result[3] = 0;
        return (TEX_INT | TEX_RGB);   
    }
    
    /* Plugin state is enabled, check if it needs to be initialized again */
    if (cast->instance_data == NULL)
    {
        MSG0("- plugin reinitialized\n");
        create_instance(cast);
    }

    Instance *instance = (Instance *) cast->instance_data;       

    /* If the file name has changed then initialize again */
    if (strcmp(instance->current_file_name, cast->file_name) != 0)
    {
        MSG0("- file name changed; plugin reinitialized\n");
        destroy_instance(cast);
        create_instance(cast);
        instance = (Instance *) cast->instance_data;       
    }
    
    /* Test for error state */
    if (instance->backend_data == NULL)
    {    
        /* Return a red pixel */
        result[0] = result[1] = result[4] = 1.0;
        result[2] = result[3] = 0;
        return (TEX_INT | TEX_RGB);
    }   

    /* If the base color has changed then empty the tile set */
    if (memcmp(instance->current_base_color, cast->base_color, 3) != 0)
    {
        MSG0("- base color changed; empty the tile set\n");
        destroy_tile_set(instance);
        create_tile_set(instance);
        memcpy(instance->current_base_color, cast->base_color, 3);
    }

    int ideal_res, power;
    if (dxt && dyt)
    {
        /* Compute pixel size by bounding box estimation
           double du = ABS(dxt[0]) + ABS(dyt[0]);
           double dv = ABS(dxt[1]) + ABS(dyt[1]);    
           double pixel_size = MAX2(du, dv); */
         
        /* Compute pixel size by area estimation */
        double signed_area = dxt[0] * dyt[1] - dxt[1] * dyt[0];
        double pixel_size = sqrt(fabs(signed_area));
        
        /* Compute the ideal image resolution and its upper bound power of two */
        ideal_res = (int)(2.0 / pixel_size);
        for (power = 1; ideal_res >> power; power++);

        /* Avoid precision errors that cause the creation of high resolution images */
        if (power > instance->tile_set_max_power)
        {
            /* The test below is faster than calculating the exact angle with
               double angle = 180.0 * atan2(signed_area, dot_product) / M_PI; */
            double dot_product = dxt[0] * dyt[0] + dxt[1] * dyt[1];
            if (dot_product != 0.0 && fabs(signed_area / dot_product) < 0.1)
            {
                /* Angle is less than 6 degrees, therefore the pixel area is very
                   thin: fallback to the max resolution used so far */ 
                MSG0("*");
                ideal_res = 1 << instance->tile_set_max_power;
                power     = instance->tile_set_max_power;
            }
        }
    }
    else
    {
        /* OSA is not active; use the default image resolution */ 
        ideal_res = TILE_SIZE;
        power     = TILE_SIZE_POWER;
    }
    int max_res = 1 << power;
    int min_res = max_res >> 1;

    /* Sample pixel from image with higher resolution */
    double r, g, b;
    sample(instance, texvec, power, cast->interpol, &r, &g, &b);
    if (cast->tex_level)
    {
        /* Apply hue color based on image detail level */
        double h = (power - 1) * 30.0;
        double cr, cg, cb;
        hsv2rgb(h, 1.0, 1.0, &cr, &cg, &cb);
        r *= cr;
        g *= cg;
        b *= cb;
    }

    /* Slight optimization: skip second image if ideal_res == max_res */ 
    if (cast->mipmap && ideal_res != max_res)
    {
        /* Sample pixel from vector image with lower resolution */
        double r2, g2, b2; 
        sample(instance, texvec, power - 1, cast->interpol, &r2, &g2, &b2);
        if (cast->tex_level)
        {
            /* Apply hue color based on texture detail level */
            double h = (power - 2) * 30.0;
            double cr, cg, cb;
            hsv2rgb(h, 1.0, 1.0, &cr, &cg, &cb);
            r2 *= cr;
            g2 *= cg;
            b2 *= cb;
        }
        
        /* Do mipmapping: interpolate final pixel from the sampled pixels */
        double max_weight = (double)(ideal_res - min_res) / (max_res - min_res);
        double min_weight = 1 - max_weight;
        r = r * max_weight + r2 * min_weight;
        g = g * max_weight + g2 * min_weight;
        b = b * max_weight + b2 * min_weight;
    }

    /* Before returning, check if the memory limit has been reached */
    if (cast->mem_max != 0 && instance->tile_memory > cast->mem_max << 20)
    {
        limit_tile_set(instance, cast->mem_max << 20);
    }
    cast->mem = (instance->tile_memory + 524288) >> 20;

    result[0] = 1.0; 
    result[1] = r;
    result[2] = g;
    result[3] = b;
    result[4] = 1.0;

    return (TEX_INT | TEX_RGB);
}
